 
/*=====================================================================*/
/* Universal plot routines for various devices                         */
/* V2.0                    by A.Vef                           08.12.93 */
/*=====================================================================*/
/* How it works:                                                       */
/*  - this module provides device-independent routines. A call to      */
/*    plot_open() connects them to the device-specific implementation  */
/*  - Method: pointers to the functions                                */
/* How to add a new driver:                                            */
/*  - define an appropriate constant PLOT_DEV_xxx in U_PL_DEF.H        */
/*  - write the driver according to the examples.                      */
/*     = all routines should be named plotd_<name>_<devtype>(),        */
/*       e.g. plotd_clear_tek()                                        */
/*     = you have to provide ALL routines, even if they are meaning-   */
/*       less for this device (e.g. input on plotter), except the      */
/*       following:                                                    */
/*     = Optional Routines:                                            */
/*         plotd_getpointer_...()                                      */
/*         plotd_checkevent_...()                                      */
/*       their presence is checked at runtime, if they do not exist,   */
/*       the calls are ignored.                                        */
/*         plotd_polysymbol_...()                                      */
/*       checked at runtime, if not present, symbols are plotted using */
/*       plotd_line_...().                                             */
/*  - add the call to plotd_init_<devtype>() to plot_open()            */
/*    if the device is limited to a specific environment, use "#ifdef" */
/*=====================================================================*/

#include "u_pl_loc.h"
#include "u_plot.h"
#include "v_inp.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


static handleptr handles[MAXHANDLES];
static int status;
static int symbols_loaded = 0;

static char plotmessages[14][40] = {"ok",
                            "No more handles",
                            "Out of memory",
                            "No such handle",
                            "No more handles of this type",
                            "Can't open Display",
                            "No such device",
                            "Error opening outputfile",
                            "file MODEL.SYM not found",
			    "Graphic Driver not found",
			    "file PLOT.FONT not found",
			    "files MODEL.SYM and PLOT.FONT not found",
                            " exit"};

/*=========== Init/Exit plot package ===============*/

#define SYMFILE "MODEL.SYM"
#define MAXSYMBOLS 34

#define FONTSFILE "PLOT.FONT"
#define SYMENV  "EFFIUTPATH"
static int fonts_loaded = 0;
char  FontName[5][80];

static struct symboltyp symbols[MAXSYMBOLS];

void load_font(){
  int i,j,k;
   char *tmp,fname[80],instr[80];
   float out[16];
   FILE *fontfile;

   if (fonts_loaded) return;
   if ((tmp=getenv(SYMENV))!=NULL){ strcpy(fname,tmp);strcat(fname,"/");
   strcat(fname,FONTSFILE);}
   else strcpy(fname,FONTSFILE);
    printf("FONTSFILE: %s\n",fname);
   if ((fontfile=fopen(fname,"r"))==NULL) return;
   for (i=0;i<5;i++){
    fgets(instr,256,fontfile);
    k=strlen(instr); instr[k-1]=instr[k];
    instr[k+1]='\0';
    printf("(%d) %s\n",i+1,instr);
    strcpy(FontName[i],instr);
    /*printf("bl[0]=%d i=%d l=%d ,%s \n",FontName[i][k-1],i,k,FontName[i]);*/
   }
   fclose(fontfile);
   fonts_loaded = TRUE;
}


static void load_sym(){
   int i,j;
   char *tmp,fname[80],instr[80];
   double out[16];
   FILE *symfile;
   if (symbols_loaded) return;
   if ((tmp=getenv(SYMENV))!=NULL) strcpy(fname,tmp);
   else strcpy(fname,SYMFILE);
   /*printf("file name=%s\n",fname);*/
   if (!(symfile=fopen(fname,"r"))) return;
   for (i=1;i<MAXSYMBOLS;i++){
      fgets(instr,70,symfile);
      trans(instr,out); symbols[i].vec=out[0];
      symbols[i].x1 = malloc(symbols[i].vec*sizeof(symbolcoordtyp));
      symbols[i].x2 = malloc(symbols[i].vec*sizeof(symbolcoordtyp));
      symbols[i].y1 = malloc(symbols[i].vec*sizeof(symbolcoordtyp));
      symbols[i].y2 = malloc(symbols[i].vec*sizeof(symbolcoordtyp));
      for (j=0; j<symbols[i].vec; j++)
	{ fgets(instr,70,symfile);
          trans(instr,out);
	  symbols[i].x1[j] = out[0];
	  symbols[i].y1[j] = out[1];
	  symbols[i].x2[j] = out[2];
	  symbols[i].y2[j] = out[3];
	}
   }
   fclose(symfile);
   symbols_loaded = TRUE;
}

int plot_init(){
   int i;
   for (i=0; i<MAXHANDLES; i++) handles[i] = NULL;
   load_sym();
   load_font();
   if (symbols_loaded) status = PLOT_STAT_OK;
                 else  status = PLOT_STAT_NOSYMBOLS;
   /*printf("sym status=%d \n",status);*/
   if (fonts_loaded)   status = PLOT_STAT_OK;
                  else status = status+PLOT_STAT_NOFONTS;
   if(status>10) status=11;
   /*printf("status=%d \n",status);*/
   return(status);
}

void plot_exit(){
   int i;
   for (i=0; i<MAXHANDLES; i++) if (handles[i]) plot_close(i);
}


static void calculate_transformation(int handle){
   handleptr h;
   h = handles[handle];
#define scale_one(x1, x2, i1, i2, f, o){\
   int i; double d;\
   i = i2 - i1;     d = x2 - x1;\
   f = d/i;         o = x1 - f * i1;}
   scale_one(h->xwmin,h->xwmax,h->xvmin,h->xvmax,h->xfact,h->xoffs);
   scale_one(h->ywmin,h->ywmax,h->yvmin,h->yvmax,h->yfact,h->yoffs);
#undef scale_one
}


/* convert screen to world coordinates */
void screen_to_world(int handle, int xi, int yi, double *xr, double *yr)
{
   *xr = handles[handle]->xfact * xi + handles[handle]->xoffs;
   *yr = handles[handle]->yfact * yi + handles[handle]->yoffs;
}

/* convert world to screen coordinates */
void world_to_screen(int handle, double xr, double yr, int *xi, int *yi)
{
   *xi = (xr - handles[handle]->xoffs) / handles[handle]->xfact;
   *yi = (yr - handles[handle]->yoffs) / handles[handle]->yfact;
}

/* Open a plot handle, i.e. Window, Plotterfile,... */
/* Returns -1 on error, otherwise handle             */
/* Name is Windowtitle or filename for Plotter      */
int plot_open(int devicetyp, int sizex, int sizey, int posx, int posy,
   	      char name[], void *special){
   int handle;
   if (devicetyp==PLOT_DEV_DEFAULT) devicetyp = PLOT_DEV_TEK;
   if (devicetyp==PLOT_DEV_DEFAULT) devicetyp = PLOT_DEV_NO;
   if ((devicetyp<=PLOT_DEV_NO) || (devicetyp>=PLOT_DEV_LASTDEV)) return(-1);
   handle = 0;
   while ((handles[handle]) && (handle <MAXHANDLES)) handle++;
   if (handle==MAXHANDLES){ status = PLOT_STAT_NOMOREHANDLES; return(-1);}
   handles[handle] = malloc(sizeof(handletyp));
   if (handles[handle]==NULL){ status = PLOT_STAT_NOMEMORY; return(-1);}
   handles[handle]->devtype = devicetyp;
   /* Initialize optional capabilities to NULL, so we can check */
   /* if they are supported.                                    */
   handles[handle]->ploth_checkevent=NULL;
   handles[handle]->ploth_getpointer=NULL;
   handles[handle]->ploth_polysymbol=NULL;
   handles[handle]->capabilities=PLOT_CAPABILITY_NONE;		 
   handles[handle]->ploth_check_any_event=NULL;		 
   handles[handle]->ploth_get_any_event=NULL;		 
   handles[handle]->ploth_rectangle = NULL;
   handles[handle]->ploth_do_special==NULL;
   handles[handle]->cm_per_unit = 0.0;
   /*printf(" u_plot.c handle=%d  devicetyp=%d \n",handle,devicetyp);*/
   switch (devicetyp){
       case PLOT_DEV_X11   : status = plotd_init_x11(handle,handles[handle],
					    sizex,sizey,posx,posy,name); break;
       case PLOT_DEV_PS    : status = plotd_init_ps(handle,handles[handle],
					     sizex,sizey,posx,posy,name,
                                             special); break;
       default: free(handles[handle]); handles[handle] = NULL;
                status = PLOT_STAT_NOSUCHDEV; return(-1);
   }
   if (status==PLOT_STAT_OK){
      (*handles[handle]->ploth_getdevice)(handle,
				       &handles[handle]->xvmax,
				       &handles[handle]->yvmax);
      handles[handle]->xwmin=0;      handles[handle]->xwmax=1;
      handles[handle]->ywmin=0;      handles[handle]->ywmax=1;
      handles[handle]->xvmin=0;
      handles[handle]->yvmin=0;
      handles[handle]->akt_x=0;      handles[handle]->akt_y=1;
      calculate_transformation(handle);
      handles[handle]->clip=FALSE;
      handles[handle]->symbol_factor = 1;
      return(handle);
   }else{
      free(handles[handle]); handles[handle] = NULL; return(-1);
   }
}

/* return to textmode:                         */
/*  - stdin/stdout available for text output   */
/*  - graphic output prohibited                */
void plot_textmode(){
   int i;
   for (i=0; i<MAXHANDLES; i++)
     if (handles[i]) (*handles[i]->ploth_textmode)(i);
}

/* switch to graphic mode                                  */
/*  - inquire actual size for each open handle (XWindow !) */
/*  - allow graphic output                                 */
void plot_graphmode(){
   int i;
   for (i=0; i<MAXHANDLES; i++){
      if (handles[i]) (*handles[i]->ploth_graphmode)(i);
   }
}

/* return last error message */
char *plot_message(){
   return plotmessages[status];
}


/*===================== Handle =====================*/
/* Modify existing handle, returns status */
int plot_modify(int handle, int sizex, int sizey, int posx, int posy,
   	        char name[]){
   if (! handles[handle]){status=PLOT_STAT_NOSUCHHANDLE; return(FALSE);}
   status = (*handles[handle]->ploth_modify)(handle,handles[handle],
                                          sizex,sizey,posx,posy,name);
   return (status==PLOT_STAT_OK);
}


/* close a handle */
void plot_close(int handle){
   if (! handles[handle]) return;
   (*handles[handle]->ploth_close)(handle);
   free(handles[handle]); handles[handle] = 0;
}

/* Clear a handle: Clear Screen or Window, FormFeed on Plotter */
void plot_clear(int handle){
   if (! handles[handle]) return;
   (*handles[handle]->ploth_clear)(handle);
}


/*===================== Inquire Status ========================*/
/* Status of handle, or general status if handle==0 */
int plot_status(int handle){
   if (handle==0) return(status);
   if (handles[handle]) return(PLOT_STAT_OK);
   return(PLOT_STAT_NOSUCHHANDLE);
}

/* return devicetype and dimensions */
int plot_getdev(int handle, int *sizex, int *sizey){
   if (! handles[handle]) { status = PLOT_STAT_NOSUCHHANDLE;
      return(PLOT_DEV_NO);   }
   (*handles[handle]->ploth_getdevice)(handle,sizex,sizey);
   return( handles[handle]->devtype);
}

/* return devicetype and dimensions if the device has the information. */
/* if not, returns PLOT_DEV_NO and unchanged parameters                */
int plot_getdev_cm(int handle, double *x_cm, double *y_cm){
   int sizex, sizey;
   if (! handles[handle]) { status = PLOT_STAT_NOSUCHHANDLE;
      return(PLOT_DEV_NO);   }
   if ((handles[handle]->capabilities & PLOT_CAPABILITY_CM)==0) 
      return(PLOT_DEV_NO);
   if (handles[handle]->cm_per_unit == 0) return(PLOT_DEV_NO);
   (*handles[handle]->ploth_getdevice)(handle,&sizex,&sizey);
   *x_cm = sizex * handles[handle]->cm_per_unit;
   *y_cm = sizey * handles[handle]->cm_per_unit;
   return( handles[handle]->devtype);
}


unsigned long plot_capabilities(int handle, unsigned long capability){
   if (capability==PLOT_CAPABILITY_ALL) 
      return handles[handle]->capabilities;
   return ((handles[handle]->capabilities & capability) != 0);
}
   
/*====================== Modify Parameters ====================*/
/* All routines return the value actually used, which may be   */
/* different from the value desired                            */

/* Set foreground colour index */
int plot_setfg(int handle, int foreground){
  return((*handles[handle]->ploth_setfg)(handle,foreground));}

/* Set background colour index */
int plot_setbg(int handle, int background){
  return((*handles[handle]->ploth_setbg)(handle,background));}

/* Set writing mode (replace/xor/or/erase/...) */
int plot_setmode(int handle, int mode){
  return((*handles[handle]->ploth_setmode)(handle,mode));}

/* Set data window */
void plot_setwin(int handle, double xmin, double ymin,
		             double xmax, double ymax){
   handles[handle]->xwmin = xmin;   handles[handle]->xwmax = xmax;
   handles[handle]->ywmin = ymin;   handles[handle]->ywmax = ymax;
   calculate_transformation(handle);
}

/* Set data viewport */
void plot_setview(int handle, int xmin, int ymin, int xmax, int ymax){
   handles[handle]->xvmin = xmin;   handles[handle]->xvmax = xmax;
   handles[handle]->yvmin = ymin;   handles[handle]->yvmax = ymax;
   calculate_transformation(handle);
}

/* set clipping */
void plot_setclip(int handle, int clip){
   handles[handle]->clip = clip;
}

/* set symbol size */
double plot_setsymsize(int handle, double size){
   handles[handle]->symbol_factor = size;
   return(handles[handle]->symbol_factor);
}


/*====================== inquire Parameters ====================*/
/* get foreground colour index */
int plot_getfg(int handle){
   return((*handles[handle]->ploth_getfg)(handle));
}


/* get background colour index */
int plot_getbg(int handle){
   return((*handles[handle]->ploth_getbg)(handle));
}


/* get writing mode (replace/xor/or/erase/...) */
int plot_getmode(int handle){
   return((*handles[handle]->ploth_getmode)(handle));
}


/* get data window */
void plot_getwin(int handle, double *xmin, double *xmax,
		             double *ymin, double *ymax){
    *xmin = handles[handle]->xwmin;    *xmax = handles[handle]->xwmax;
    *ymin = handles[handle]->ywmin;    *ymax = handles[handle]->ywmax;
}

/* get data viewport */
void plot_getview(int handle, int *xmin, int *xmax, int *ymin, int *ymax){
    *xmin = handles[handle]->xvmin;    *xmax = handles[handle]->xvmax;
    *ymin = handles[handle]->yvmin;    *ymax = handles[handle]->yvmax;
}


/* get clipping */
int plot_getclip(int handle){return(handles[handle]->clip);
}


/* get symbol size */
double plot_getsymsize(int handle){return(handles[handle]->symbol_factor);
}


/*=========== Output routines, world coordinates only ============*/
static int in_range(int handle, double x, double y)
{ if (!handles[handle]->clip) return(1);
  if (x > handles[handle]->xwmax) return(0);
  if (y > handles[handle]->ywmax) return(0);
  if (x < handles[handle]->xwmin) return(0);
  if (y < handles[handle]->ywmin) return(0);
  return(1);
}

void plot_point(int handle, double X, double Y){
   int IX, IY;
   if (in_range(handle,X,Y)){
      world_to_screen(handle,X,Y,&IX,&IY);
      (*handles[handle]->ploth_point)(handle,X,Y);
}}

void plot_line(int handle, double X0, double Y0, double X1, double Y1){
   int IX0, IY0, IX1, IY1;
   if ((in_range(handle,X0,Y0)) && (in_range(handle,X1,Y1))){
      world_to_screen(handle,X0,Y0,&IX0,&IY0);
      world_to_screen(handle,X1,Y1,&IX1,&IY1);
      (*handles[handle]->ploth_line)(handle,IX0,IY0,IX1,IY1);
}}


/*=================== Text output ================*/
/* set text alignment */
void plot_settextalign(int handle, int xalign, int yalign){
   (*handles[handle]->ploth_settextalign)(handle,xalign,yalign);
}

/* write txt at position (X,Y) */
void plot_writetext(int handle, int worldcoor, double X, double Y, char txt[]){
   int IX,IY;
   if (worldcoor) world_to_screen(handle,X,Y,&IX,&IY);
   else { IX = X; IY = Y; }
   (*handles[handle]->ploth_writetext)(handle,IX,IY,txt);
}

/* erase txt at position (X,Y). Usually done by plot_setmode(ERASE) and */
/* plot_writetext, but in some cases it has to be done in another way...*/
void plot_erasetext(int handle, int worldcoor, double X, double Y, char txt[]){
   int IX,IY;
   if (worldcoor) world_to_screen(handle,X,Y,&IX,&IY);
   else { IX = X; IY = Y; }
   (*handles[handle]->ploth_erasetext)(handle,IX,IY,txt);
}


/*================= Pointer position ==============*/

/* Get pointer position and return terminator. Returns 0 if handle if */
/* output only (Plotter,...), character if a key was pressedm, or one */
/* of the codes defined as PLOT_TRM_... in "u_pl_def.h"               */
int plot_getpointer(int handle, double *X, double *Y){
   int IX, IY, trm;
   if (handles[handle]->ploth_getpointer==NULL) return 0;
   trm = (*handles[handle]->ploth_getpointer)(handle,&IX,&IY);
   if (trm != PLOT_TRM_ABORT) screen_to_world(handle,IX,IY,X,Y);
   return(trm);
}

/* Check for event, if event present, get pointer position and return */
/* terminator. Returns 0 if handle if output only (Plotter,...) or    */
/* no event, character if a key was pressedm, or one of the codes     */
/* defined as PLOT_TRM_... in "u_pl_def.h"                            */
int plot_checkevent(int handle, double *X, double *Y){
   int IX, IY, trm;
   if (handles[handle]->ploth_checkevent==NULL) return 0;
   trm = (*handles[handle]->ploth_checkevent)(handle,&IX,&IY);
   if (trm != PLOT_TRM_ABORT) screen_to_world(handle,IX,IY,X,Y);
   return(trm);
}

int plot_check_any_event(int handle, int *at_handle, double *X, double *Y){
   int trm, IX, IY;
   if (handles[handle]->ploth_check_any_event==NULL){ 
      *at_handle = handle; 
      return plot_checkevent(handle,X,Y); 
   }
   trm = (*handles[handle]->ploth_check_any_event)(handle,at_handle,&IX,&IY);
   if (trm != PLOT_TRM_ABORT) screen_to_world(*at_handle,IX,IY,X,Y);
   return(trm);
}

int plot_get_any_event(int handle, int *at_handle, double *X, double *Y){
   int trm, IX, IY;
   if (handles[handle]->ploth_get_any_event==NULL){ 
      printf("Not impl");newline();
      *at_handle = handle; 
      return plot_getpointer(handle,X,Y); 
   }
   trm = (*handles[handle]->ploth_get_any_event)(handle,at_handle,&IX,&IY);
   if (trm != PLOT_TRM_ABORT) screen_to_world(*at_handle,IX,IY,X,Y);
   return(trm);
}

/*==================== Future work ================*/
/* Set font for text in graphic mode */
/* Define color                      */


void plot_lineabs(int handle, double x1, double y1, double x2, double y2){
   plot_line(handle,x1,y1,x2,y2);
   handles[handle]->akt_x = x2; handles[handle]->akt_y = y2;
}


void plot_lineto(int handle, double x, double y){
   plot_lineabs(handle,handles[handle]->akt_x,
		       handles[handle]->akt_y,x,y);
}

void plot_linerel(int handle, double x, double y){
   x = handles[handle]->akt_x + x; y = handles[handle]->akt_y + y;
   plot_lineto(handle,x,y);
}

void plot_moverel(int handle, double x, double y){
   handles[handle]->akt_x = handles[handle]->akt_x + x;
   handles[handle]->akt_y = handles[handle]->akt_y + y;
}

void plot_moveto(int handle, double x, double y){
   handles[handle]->akt_x = x;
   handles[handle]->akt_y = y;
}

void plot_polyline(int handle, int n, double x[], double y[]){
   int *ix, *iy, i, j, size; 
   size = n*sizeof(int); ix = malloc(size); iy = malloc(size);
   if ((ix==NULL) || (iy==NULL)){
      free(ix); free(iy);
      status = PLOT_STAT_NOMEMORY; return;}
   j = 0;
   for (i=0; i<n; i++) {
      if (in_range(handle,x[i],y[i])){
        world_to_screen(handle,x[i],y[i],ix+j,iy+j); j++;
      } else {
      if (j>1) (*handles[handle]->ploth_polyline)(handle,j,ix,iy); j=0;}
   }
   if (j>1) (*handles[handle]->ploth_polyline)(handle,j,ix,iy);
   free(ix); free(iy);
}
void plot_fillpolygon(int handle, int n, double x[], double y[]){
   int *ix, *iy, i, j, size; 
   size = n*sizeof(int); ix = malloc(size); iy = malloc(size);
   if ((ix==NULL) || (iy==NULL)){
      free(ix); free(iy);
      status = PLOT_STAT_NOMEMORY; return;}
   j = 0;
   for (i=0; i<n; i++) {
      if (in_range(handle,x[i],y[i])){
        world_to_screen(handle,x[i],y[i],ix+j,iy+j); j++;
      } else {
      if (j>1) (*handles[handle]->ploth_fillpolygon)(handle,j,ix,iy); j=0;}
   }
   if (j>1) (*handles[handle]->ploth_fillpolygon)(handle,j,ix,iy);
   free(ix); free(iy);
}

void plot_rectangle(int handle, double x1, double y1, 
		    double x2, double y2, int fillcolor){
  int ix1,iy1,ix2,iy2, *x, *y, savecol;
  if (! in_range(handle,x1,y1)) return;
  if (! in_range(handle,x2,y2)) return;
  world_to_screen(handle,x1,y1,&ix1,&iy1);
  world_to_screen(handle,x2,y2,&ix2,&iy2);
  if (handles[handle]->ploth_rectangle==NULL){
    savecol  = plot_getfg(handle);  
    plot_setfg(handle,fillcolor);
    x = malloc(5*sizeof(int));
    y = malloc(5*sizeof(int));
    x[0] = x[3] = x[4] = ix1; x[1] = x[2] = ix2;			    
    y[0] = y[1] = y[4] = iy1; y[2] = y[3] = iy2;			    
    (*handles[handle]->ploth_polyline)(handle,5,x,y);
    free(x); free(y);
    plot_setfg(handle,savecol);
  }else	{
    (*handles[handle]->ploth_rectangle)(handle,ix1,iy1,ix2,iy2,fillcolor);
  }
}
		       
		       
void plot_symbol(int handle, int sym, double x, double y)
{  int i, ix, iy;
   if (! in_range(handle,x,y)) return;
   world_to_screen(handle,x,y, &ix, &iy);
   if ((symbols_loaded) && (sym>0) && (sym<34)){
     for (i=0; i<symbols[sym].vec; i++)
	(*handles[handle]->ploth_line)(handle,
	   (int)(ix+symbols[sym].x1[i] * handles[handle]->symbol_factor),
	   (int)(iy+symbols[sym].y1[i] * handles[handle]->symbol_factor),
	   (int)(ix+symbols[sym].x2[i] * handles[handle]->symbol_factor),
	   (int)(iy+symbols[sym].y2[i] * handles[handle]->symbol_factor));
   } else (*handles[handle]->ploth_point)(handle,ix,iy);
}

void plot_polysymbol(int handle, int pts, int sym, double X[], double Y[]){
   int i, *ix, *iy, ipts;
   if ((symbols_loaded) && (sym>0) && (sym<34)){
     if (handles[handle]->ploth_polysymbol==NULL)
       for (i=0; i<pts; i++) plot_symbol(handle,sym,X[i],Y[i]);
     else{
       ix = malloc(pts * sizeof(int));
       iy = malloc(pts * sizeof(int));
       ipts = 0;
       for (i=0; i<pts; i++) {
         if (! in_range(handle,X[i],Y[i])) continue;
         world_to_screen(handle,X[i],Y[i], &ix[ipts], &iy[ipts]);
	 ipts++;
       }
       (*handles[handle]->ploth_polysymbol)(handle,ipts,symbols[sym],ix,iy);
       free(ix); free(iy);
   }} else for (i=0; i<pts; i++) plot_point(handle,X[i],Y[i]);
}

/* Write a string transparently to the device, if appropriate. */
/* The device driver may ignore it.                            */
void plot_do_special(int handle, char str[]){
  if (handles[handle]->ploth_do_special==NULL) return;
  (*handles[handle]->ploth_do_special)(handle,str);
}
