
#include "u_pl_def.h"

/*=========== Init/Exit plot package =================*/

/* initialize, load symbols. Returns FALSE if symbols */
/* were not found, TRUE otherwise                     */
int plot_init(void);

/* close all handles and exit */
void plot_exit(void);

/* Open a plot handle, i.e. Window, Plotterfile,... */
/* Returns 0 on error, otherwise handle             */
/* Name is Windowtitle or filename for Plotter      */
/* Special is a pointer to a device dependent data structure, or NULL */
/*   Usage of Special:                                                */
/*   - PostScript : int[4] : Bounding Box as xmin, ymin, xmax, ymax   */

int plot_open(int devicetyp, int sizex, int sizey, int posx, int posy,
   	      char name[], void *special);


/* return to textmode:                         */
/*  - stdin/stdout available for text output   */
/*  - graphic output prohibited                */
void plot_textmode(void);

/* switch to graphic mode                                  */
/*  - inquire actual size for each open handle (XWindow !) */
/*  - allow graphic output                                 */
void plot_graphmode(void);

/* return last error message */
char *plot_message(void);

/*===================== Handle =====================*/
/* Modify existing handle, returns status */
int plot_modify(int handle, int sizex, int sizey, int posx, int posy,
   	        char name[]);

/* close a handle */
void plot_close(int handle);

/* Clear a handle: Clear Screen or Window, FormFeed on Plotter */
void plot_clear(int handle);

/*===================== Inquire Status ========================*/
/* Status of handle, or general status if handle==0 */
int plot_status(int handle);

/* return devicetype and dimensions */
int plot_getdev(int handle, int *sizex, int *sizey);

/* return devicetype and dimensions in cm if the device has this    */
/* information. If not, return PLOT_DEV_NO and unchanged parameters */
int plot_getdev_cm(int handle, double *x_cm, double *y_cm);

/* return capabilities of the device as a longword bitmap.  */
/* To inquire a special capability, pass its value, in this */
/* case, TRUE is return if the capability is present        */
unsigned long plot_capabilities(int handle, unsigned long capability);

/*====================== Modify Parameters ====================*/
/* All routines return the value actually used, which may be   */
/* different from the value desired                            */

/* Set foreground colour index */
int plot_setfg(int handle, int foreground);

/* Set background colour index */
int plot_setbg(int handle, int background);

/* Set writing mode (replace/xor/or/erase/...) */
int plot_setmode(int handle, int mode);

/* Set data window */
void plot_setwin(int handle, double xmin, double ymin,
		             double xmax, double ymax);

/* Set data viewport */
void plot_setview(int handle, int xmin, int ymin, int xmax, int ymax);

/* set clipping */
void plot_setclip(int handle, int clip);

/* set symbol size */
double plot_setsymsize(int handle, double size);

/*====================== inquire Parameters ====================*/
/* get foreground colour index */
int plot_getfg(int handle);

/* get background colour index */
int plot_getbg(int handle);

/* get writing mode (replace/xor/or/erase/...) */
int plot_getmode(int handle);

/* get data window */
void plot_getwin(int handle, double *xmin, double *xmax,
		             double *ymin, double *ymax);

/* get data viewport */
void plot_getview(int handle, int *xmin, int *xmax, int *ymin, int *ymax);

/* get clipping */
int plot_getclip(int handle);

/* get symbol size */
double plot_getsymsize(int handle);

/*=========== Output routines, world coordinates only ============*/

void plot_point(int handle, double X, double Y);

void plot_symbol(int handle, int sym, double X, double Y);

void plot_polysymbol(int handle, int pts, int sym, double X[], double Y[]);

void plot_line(int handle, double X0, double Y0, double X1, double Y1);

void plot_polyline(int handle, int pts, double X[], double Y[]);

/*=================== Text output ================*/
/* Set text alignment */
void plot_settextalign(int handle, int xalign, int yalign);

/* write txt at position (X,Y) */
void plot_writetext(int handle, int worldcoor, double X, double Y, char txt[]);

/* erase txt at position (X,Y). Usually done by plot_setmode(ERASE) and */
/* plot_writetext, but in some cases it has to be done in another way...*/
void plot_erasetext(int handle, int worldcoor, double X, double Y, char txt[]);

/*================= Pointer position ==============*/

/* Get pointer position and return terminator. Returns PLOT_TRM_ABORT */
/* if handle is output only (Plotter,...), character if a key was     */
/* pressed, or one of the PLOT_TRM_... codes defined in "u_pl_def.h"  */
int plot_getpointer(int handle, double *X, double *Y);

/* Check for event, if event present, get pointer position and return */
/* terminator. Returns PLOT_TRM_ABORT if handle is output only        */
/* (Plotter,...), cannot check for events, or has no pending event.   */
/* Otherwise, character if a key was pressed, or one of the codes     */
/* defined as PLOT_TRM_... in "u_pl_def.h"                            */
int plot_checkevent(int handle, double *X, double *Y);

/* <plot_get_any_event> and <plot_check_any_event> are similar to     */
/* <plot_getpointer> and <plot_checkevent>, but they check any device */
/* of the same type as <handle>. The handle of the device in which    */
/* the event occured is returned in <at_handle>.                      */
/* They can be used on all devices. If they are not present, the call */
/* is forwarded to <plot_getpointer> and <plot_checkevent>            */
/* Only on window systems you can have more than one handle of a type */

int plot_get_any_event(int handle, int *at_handle, double *X, double *Y);
int plot_check_any_event(int handle, int *at_handle, double *X, double *Y);

/*======== Higher level routines: relative lines etc. =======*/
/* draw line from (X1,Y1) to (X2,Y2) and set actual position to (X2,Y2) */ 
void plot_lineabs(int handle, double x1, double y1, double x2, double y2);
/* draw line from actual position to (X,Y) and set actual position to (X,Y) */ 
void plot_lineto(int handle, double x, double y);
/* draw line from actual position to a point in the distance (X,Y) and */
/* set actual position to this point                                   */ 
void plot_linerel(int handle, double x, double y);
/* move actual position by (X,Y) */
void plot_moverel(int handle, double x, double y);
/* move actual position to (X,Y) */
void plot_moveto(int handle, double x, double y);
/* Draw a rectangle with corners (X1,Y1) and (X2,Y2). If the device */
/* supports this operation, the rectangle is filled unsing color    */
/* fillcolor.                                                       */
void plot_rectangle(int handle, double x1, double y1, 
                    double x2, double y2, int fillcolor);

/* Transform between World(User) and Screen(Device) coordinates */
void screen_to_world(int handle, int xi, int yi, double *xr, double *yr);
void world_to_screen(int handle, double xr, double yr, int *xi, int *yi);

/* Write a string transparently to the device, if appropriate. */
/* The device driver may ignore it.                            */
void plot_do_special(int handle, char str[]);

/*==================== Future work ================*/ 
/* Set font for text in graphic mode */
/* Define color                      */

